"""
Dettectinator - The Python library to your DeTT&CT YAML files.
Authors:
    Martijn Veken, Sirius Security
    Ruben Bouman, Sirius Security
License: GPL-3.0 License
"""

import json
import sys
import os
import importlib
import inspect

from dettectinator import DettectTechniquesAdministration, DettectDataSourcesAdministration
from plugins.technique_import import TechniqueBase
from plugins.datasources_import import DatasourceBase
from argparse import ArgumentParser, Namespace


class CommandLine:

    @staticmethod
    def _get_raw_commandline(argument_names: list) -> str:
        """
        Some arguments need to be read from the command line before being processed with ArgumentParser.
        This function provides a way to read these values from the command line in a simple way.
        It's ugly, but it works :)
        :return: The value of the requested argument.
        """
        prev = ''
        value = ''
        for item in sys.argv:
            if prev in argument_names:
                value = item
                break
            prev = item
        return value

    @staticmethod
    def _print_plugins(import_plugins: dict) -> None:
        """
        Prints the list of available data import plugins in a module
        :param import_plugins: dictionary containing the plugins
        """
        for name in import_plugins.keys():
            print(f' - {name}')

    @staticmethod
    def _get_plugins() -> dict:
        """
        Retrieves all plugins from the plugin folder
        :return: dict containing plugins and modules
        """
        import_plugins = {}
        path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'plugins')
        for module in [x for x in os.listdir(path) if x[-3:] == '.py']:
            plugin_mod = importlib.import_module('plugins.' + module[:-3])
            for name, cls in inspect.getmembers(plugin_mod, inspect.isclass):
                if ('Technique' in name or 'Datasource' in name) and 'Base' not in name:
                    import_plugins[name] = plugin_mod
        return import_plugins

    @staticmethod
    def _set_default_params(parser: ArgumentParser):
        """
        Set the default command line arguments
        """
        required = parser.add_argument_group('required arguments')
        parser.add_argument('-c', '--config', help='Configuration file location.')
        required.add_argument('-p', '--plugin', help='Data import plugin name.', required=True)
        required.add_argument('-a', '--applicable_to',
                              help='Systems that the detections are applicable to (comma seperated list).',
                              required=True)
        parser.add_argument('-d', '--domain',
                            help='The ATT&CK domain (default = enterprise). This argument is ignored if a domain is specified in the YAML file.',
                            required=False, choices=['enterprise', 'ics', 'mobile'])
        parser.add_argument('-i', '--input_file', help='YAML filename for input.', default=None)
        parser.add_argument('-o', '--output_file', help='YAML filename for output.', default=None)
        parser.add_argument('-n', '--name', help='Value for the name attribute in the YAML file.', default=None)
        parser.add_argument('-s', '--stix_location', help='Local STIX repository location.', default=None)

        parser.add_argument('-ch', '--check_unused', action='store_true', help='Check unused detections.')
        parser.add_argument('-cl', '--clean_unused', action='store_true', help='Clean unused detections.')

    def _get_argument_values_from_config_file(self) -> dict:
        """
        Read the command line arguments from the config file if applicable
        """
        config_file_name = self._get_raw_commandline(['-c', '--config'])
        if config_file_name:
            print(f'Reading settings from "{config_file_name}".')
            with open(config_file_name, 'r') as f:
                config_file_arguments = json.load(f)
        else:
            config_file_arguments = {}
        return config_file_arguments

    @staticmethod
    def process_techniques(applicable_to: list, arguments: Namespace, plugin: TechniqueBase) -> tuple:
        """
        Process all techniques from the source system
        """
        # Get the technique data
        techniques = plugin.get_attack_techniques(applicable_to)
        # Convert data to yaml
        print('Generating techniques YAML file.')
        dettect = DettectTechniquesAdministration(arguments.input_file, domain=arguments.domain,
                                                  local_stix_path=arguments.stix_location)

        location_prefix_unused_detections = arguments.location_prefix if arguments.clean_unused_location_prefix else ''

        warnings, results = dettect.update_detections(techniques, check_unused_detections=arguments.check_unused,
                                                      clean_unused_detections=arguments.clean_unused,
                                                      location_prefix_unused_detections=location_prefix_unused_detections)
        return dettect, results, warnings

    @staticmethod
    def process_datasource(applicable_to: list, arguments: Namespace, plugin: DatasourceBase) -> tuple:
        """
        Process all data sources from the source system
        """
        # Get the data source data
        datasources = plugin.get_attack_datasources(applicable_to)
        # Convert data to yaml
        print('Generating datasources YAML file.')
        dettect = DettectDataSourcesAdministration(arguments.input_file, domain=arguments.domain,
                                                   local_stix_path=arguments.stix_location)
        warnings, results = dettect.update_data_sources(datasources, check_unused_data_sources=arguments.check_unused,
                                                        clean_unused_data_sources=arguments.clean_unused)
        return dettect, results, warnings

    def start(self) -> None:
        """
        Dettectinator has been started from the commandline.
        Process the command line arguments and launch the appropriate plugin.
        """

        # Load default argument values from the config file
        config_file_arguments = self._get_argument_values_from_config_file()

        # Retrieve all available plugins
        plugins = self._get_plugins()

        # Get the plugin name from the arguments
        plugin_name = config_file_arguments.get('plugin', self._get_raw_commandline(['-p', '--plugin']))

        if plugin_name:
            # Get the plugin class if it exists
            if plugin_name in plugins.keys():
                plugin_class = getattr(plugins[plugin_name], plugin_name)
                print(f'Plugin "{plugin_name}" has been found.')
            else:
                print(f'data import plugin "{plugin_name}" does not exist. Valid plugins:')
                self._print_plugins(plugins)
                sys.exit()

            # Add the default command line params
            parser = ArgumentParser(add_help=True, conflict_handler='error', )
            self._set_default_params(parser)

            # Add the parameters from the plugin
            plugin_group = parser.add_argument_group(plugin_name)
            plugin_class.set_plugin_params(plugin_group)

            # Set the default values from the config file
            # Default and required don't work together, so set required to False
            parser.set_defaults(**config_file_arguments)
            for action in parser._actions:
                if action.dest in config_file_arguments.keys():
                    action.required = False

            # Evaluate command line arguments
            arguments = parser.parse_args()
            applicable_to = [at.strip() for at in arguments.applicable_to.split(',')]
            output_file = arguments.output_file or arguments.input_file

            # Read the data from the source
            print(f'Using "{plugin_name}" to collect data.')
            plugin = plugin_class(vars(arguments))

            if plugin_name.startswith('Technique'):
                dettect, results, warnings = self.process_techniques(applicable_to, arguments, plugin)
            else:
                dettect, results, warnings = self.process_datasource(applicable_to, arguments, plugin)

            if arguments.name:
                dettect.set_name(arguments.name)

            dettect.save_yaml_file(output_file)
            print(f'DeTT&CT YAML file written: {dettect.get_filename()}')

            output = warnings + results
            if len(output) > 0:
                print('\nPlease review the following items:')
                print(' - ' + '\n - '.join(output))
        else:
            print('Please specify a valid data import plugin using the "-p" argument:')
            self._print_plugins(plugins)



"""
Dettectinator - The Python library to your DeTT&CT YAML files.
Authors:
    Martijn Veken, Sirius Security
    Ruben Bouman, Sirius Security
License: GPL-3.0 License

Some functions and code are taken with permission from the DeTT&CT project (http://github.com/rabobank-cdc/DeTTECT).
"""

import importlib
import inspect
import sys
import os
import json
import errno
from io import StringIO
from datetime import datetime
from copy import deepcopy
from logging import getLogger, ERROR as LOGERROR
from ruamel.yaml import YAML
from attackcti import attack_client
from requests import exceptions
from stix2 import datastore, Filter
from anyascii import anyascii
import dateutil.parser


getLogger('taxii2client').setLevel(LOGERROR)

try:
    # When dettectinator is installed as python library
    from dettectinator.constants import *
except ModuleNotFoundError:
    # When dettectinator is not installed as python library
    sys.path.append(os.path.dirname(os.path.abspath(__file__)))
    from constants import *


class DettectBase(object):
    """
    Base class for DeTT&CT YAML file classes.
    """

    def __init__(self, filename: str, domain: str = None, local_stix_path: str = None) -> None:
        self.filename = filename
        self._local_stix_path = local_stix_path
        self._yaml = None
        self._yaml_content = None

        self._init_yaml()

        if filename is not None:
            self._load_yaml_content()
            self.domain = 'enterprise-attack' if 'domain' not in self._yaml_content.keys() else self._yaml_content['domain']
        else:
            self.domain = 'mobile-attack' if domain == 'mobile' else 'ics-attack' if domain == 'ics' else 'enterprise-attack'
            self.start_clean_file()

        self._load_platform_in_correct_capitalisation()
        self.name = self._yaml_content['name']
        self._initialize_attack_client()
        self._load_attack_techniques()

        # TODO YAML merge functionaliteit

    def _init_yaml(self) -> None:
        """
        Initialize ruamel.yaml with the correct settings
        """
        self._yaml = YAML()
        self._yaml.Representer.ignore_aliases = lambda *args: True  # disable anchors/aliases

    def _load_yaml_content(self) -> None:
        """
        Loads the YAML file into _yaml_content. It will contain the raw YAML content. This variable is used for alteration of the YAML file.
        """
        if not os.path.exists(self.filename):
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self.filename)

        with open(self.filename, 'r') as yaml_file:
            self._yaml_content = self._yaml.load(yaml_file)

    def set_name(self, name):
        """
        Sets the name attribute in the YAML file.
        """
        self._yaml_content['name'] = name

    def get_filename(self):
        """
        Returns the filename of the YAML file.
        """
        return self.filename

    def start_clean_file(self):
        """
        Starts a clean YAML file.
        """
        if isinstance(self, DettectDataSourcesAdministration):
            self._yaml_content = deepcopy(YAML_OBJ_NEW_DATA_SOURCES_FILE)
            self.filename = f"data_sources_{self._yaml_content['name']}.yaml"
            self._yaml_content['domain'] = self.domain

            if self.domain == 'mobile-attack':
                raise Exception(f'{self.__class__.__name__}: Data sources for ATT&CK Mobile not yet implemented by MITRE.')
        elif isinstance(self, DettectTechniquesAdministration):
            self._yaml_content = deepcopy(YAML_OBJ_NEW_TECHNIQUES_FILE)
            self.filename = f"techniques_{self._yaml_content['name']}.yaml"
            self._yaml_content['domain'] = self.domain

    def save_yaml_file(self, filename=None) -> None:
        """
        Saves the YAML file back to disk using the given filename or the same filename when the file was read.
        """
        with open(filename if filename else self.filename, 'w') as yaml_file:
            string_io = StringIO()
            self._yaml.indent(mapping=4, sequence=4, offset=2)
            self._yaml.width = 2048
            self._yaml.dump(self._yaml_content, string_io)
            string_io.seek(0)
            new_lines = string_io.readlines()
            for line in new_lines:
                # Date fix for compatability with YAML Editor (which uses JavaScript to generate YAML):
                if 'date' in line:
                    line = line.replace('T00:00:00Z', 'T00:00:00.000Z').replace(' 00:00:00', 'T00:00:00.000Z')
                yaml_file.write(anyascii(line))
            self.filename = filename if filename else self.filename

    @staticmethod
    def _set_yaml_dv_comments(yaml_object: object) -> object:
        """
        Set all comments in the detection, visibility or data source details YAML object when the 'comment' key-value pair is missing or is None.
        This gives the user the flexibility to have YAML files with missing 'comment' key-value pairs.
        :param yaml_object: detection or visibility object.
        :return: detection or visibility object for which empty comments are filled with an empty string.
        """
        yaml_object['comment'] = yaml_object.get('comment', '')
        if yaml_object['comment'] is None:
            yaml_object['comment'] = ''
        if 'score_logbook' in yaml_object:
            for score_obj in yaml_object['score_logbook']:
                score_obj['comment'] = score_obj.get('comment', '')
                if score_obj['comment'] is None:
                    score_obj['comment'] = ''

        return yaml_object

    @staticmethod
    def _add_entry_to_list_in_dictionary(dictionary: dict, key_dict: str, key_list: str, entry: object) -> None:
        """
        Ensures a list will be created if it doesn't exist in the given dict[key_dict][key_list] and adds the entry to the
        list. If the dict[key_dict] doesn't exist yet, it will be created.
        :param dictionary: the dictionary
        :param key_dict: the key name in de main dict
        :param key_list: the key name where the list in the dictionary resides
        :param entry: the entry to add to the list
        :return:
        """
        if key_dict not in dictionary.keys():
            dictionary[key_dict] = {}
        if key_list not in dictionary[key_dict].keys():
            dictionary[key_dict][key_list] = []
        dictionary[key_dict][key_list].append(entry)

    def _load_platform_in_correct_capitalisation(self) -> None:
        """
        Sets the platforms with the correct capitalisation. E.g.: "linux" will become "Linux"
        """
        platform = self._yaml_content.get('platform', None)

        if platform is None:
            platform = []

        if isinstance(platform, str):
            platform = [platform]

        platform = [p.lower() for p in platform if p is not None]
        selected_platforms = PLATFORMS_ENTERPRISE if self.domain == 'enterprise-attack' else PLATFORMS_ICS if self.domain == 'ics-attack' else PLATFORMS_MOBILE

        if 'all' in platform:
            platform = list(selected_platforms.values())
        else:
            platform = [selected_platforms[p] for p in platform if p is not None if p in selected_platforms.keys()]

        self.platform = platform

    def _initialize_attack_client(self) -> None:
        """
        Initializes the attack_client to get information from local STIX repository or from online TAXII server.
        """
        try:
            if self._local_stix_path is not None:
                if self._local_stix_path is not None and os.path.isdir(os.path.join(self._local_stix_path, 'enterprise-attack')) \
                        and os.path.isdir(os.path.join(self._local_stix_path, 'ics-attack')) \
                        and os.path.isdir(os.path.join(self._local_stix_path, 'mobile-attack')):
                    self.mitre = attack_client(local_path=self._local_stix_path)
                else:
                    raise Exception(f'{self.__class__.__name__}: Not a valid local STIX path: {self._local_stix_path}')
            else:
                self.mitre = attack_client()
        except (exceptions.ConnectionError, datastore.DataSourceError) as e:
            if hasattr(e, 'request'):
                raise Exception(f'{self.__class__.__name__}: Cannot connect to MITRE\'s CTI TAXII server: {str(e.request.url)}') from e
            else:
                raise Exception(f'{self.__class__.__name__}: Cannot connect to MITRE\'s CTI TAXII server.') from e

    def _load_attack_techniques(self) -> None:
        """
        Load the ATT&CK STIX data from the MITRE TAXII server.
        :return: MITRE ATT&CK data object (STIX)
        """
        if self.domain == 'enterprise-attack':
            self.attack_techniques = self._convert_stix_techniques_to_dict(self.mitre.get_enterprise_techniques())
        elif self.domain == 'ics-attack':
            self.attack_techniques = self._convert_stix_techniques_to_dict(self.mitre.get_ics_techniques())
        else:
            self.attack_techniques = self._convert_stix_techniques_to_dict(self.mitre.get_mobile_techniques())

    def _get_technique_from_attack(self, technique_id: str) -> dict:
        """
        Return the technique object for the given technique_id.
        """
        for tech in self.attack_techniques:
            if tech['technique_id'] == technique_id:
                return tech
        return None

    @staticmethod
    def _convert_stix_techniques_to_dict(stix_attack_data: list) -> list:
        """
        Convert the STIX list with AttackPatterns to a dictionary for easier use in python and also include the technique_id and DeTT&CT data sources.
        :param stix_attack_data: the MITRE ATT&CK STIX dataset with techniques
        :return: list with dictionaries containing all techniques from the input stix_attack_data
        """
        attack_data = []
        for stix_tech in stix_attack_data:
            tech = json.loads(stix_tech.serialize(), object_hook=DettectBase._date_hook)

            # Add technique_id as key, because it's hard to get from STIX:
            tech['technique_id'] = DettectBase._get_attack_id(stix_tech)

            # Create empty x_mitre_data_sources key for techniques without data sources:
            if 'x_mitre_data_sources' not in tech.keys():
                tech['x_mitre_data_sources'] = []

            attack_data.append(tech)

        return attack_data

    @staticmethod
    def _date_hook(json_dict: dict) -> dict:
        """
        Parses STIX dates so that they can be used as date object in dictionaries. Function is used as object_hook function in the JSON serialize.
        :param json_dict: the dictionary with STIX data
        :return: dictionary with corrected STIX data
        """
        for (key, value) in json_dict.items():
            if key == 'created':
                json_dict['created'] = dateutil.parser.parse(value)
            elif key == 'modified':
                json_dict['modified'] = dateutil.parser.parse(value)
        return json_dict

    @staticmethod
    def _get_attack_id(stix_obj: object) -> str:
        """
        Get the Technique, Group or Software ID from the STIX object
        :param stix_obj: STIX object (Technique, Software or Group)
        :return: ATT&CK ID
        """
        for ext_ref in stix_obj['external_references']:
            if ext_ref['source_name'] in ['mitre-attack', 'mitre-mobile-attack', 'mitre-ics-attack']:
                return ext_ref['external_id']


class DettectTechniquesAdministration(DettectBase):
    """
    Create or modify a DeTT&CT techniques administration YAML file.
    """

    def __init__(self, filename: str = None, domain: str = None, local_stix_path: str = None) -> None:
        super(DettectTechniquesAdministration, self).__init__(filename, domain, local_stix_path)

        # Get the normalized content from the YAML file:
        self._load_techniques()

    def update_detections(self, detection_rules: dict, check_unused_detections: bool = False, clean_unused_detections: bool = False, location_prefix_unused_detections: str = '', check_unused_applicable_to: bool = False, clean_unused_applicable_to: bool = False) -> list:
        """
        Updates the techniques YAML file with the given detections.
        :param  detection_rules: a dictionary of dictionaries containing detection name as key with the following data: applicable_to, location prefix and list with ATT&CK (sub) techniques
                {
                    {'Detection A': {'applicable_to': ['all'], 'location_prefix': 'SIEM', 'techniques': ['T1055']}}
                }
        :param  check_unused_detections: boolean. When True, all detections in YAML file will be checked against given detection_rules list. It's using
                                         the location field as list for detection rules including location_prefix as prefix: "Splunk: Detection A"
        :param  clean_unused_detections: boolean. When True, all detections in YAML file that don't exist in the given detection_rules list will be removed.
        :param clean_unused_applicable_to: boolean. When True, all detection objects with a applicable_to value that doesn't exist in the given detection_rules list will be removed.
        :return a list with results containing warnings or errors during the update process.
        """
        date_today = datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)

        warnings, results = self._add_rules(detection_rules, date_today)
        w, r = self._delete_rules(detection_rules, check_unused_detections, clean_unused_detections, location_prefix_unused_detections, check_unused_applicable_to, clean_unused_applicable_to, date_today)
        warnings += w
        results += r

        return warnings, results

    def _add_rules(self, detection_rules: dict, date_today: datetime) -> list:
        """
        Adds new detection rules to the techniques YAML file. A score_logbook entry is added for every change.
        """
        warnings, results = [], []

        # Loop through all detection rules:
        for rule_name, rule_data in detection_rules.items():
            # Loop through every technique per rule:
            for technique_id in rule_data['techniques']:
                attack_technique = self._get_technique_from_attack(technique_id)
                if attack_technique is None:
                    warnings.append(f'Technique "{technique_id}" listed in detection rule "{rule_name}" does not exist in ATT&CK ({self.domain}). Skipping.')
                    continue

                location = rule_name if rule_data['location_prefix'] == '' else rule_data['location_prefix'] + ': ' + rule_name
                yaml_technique = self._get_technique_from_yaml(technique_id)
                if yaml_technique is not None:
                    # Check is applicable_to is already there:
                    applicable_to_list = [d['applicable_to'] for d in yaml_technique['detection']]
                    if rule_data['applicable_to'] in applicable_to_list:
                        # applicable_to already present, go to the right applicable_to:
                        for d in yaml_technique['detection']:
                            if d['applicable_to'] == rule_data['applicable_to']:
                                # Check if detection rule is in location field:
                                rule_exist = False
                                for loc in d['location']:
                                    if rule_name in loc:
                                        rule_exist = True
                                        break

                                # If detection rule is not yet in location field, add detection rule to location field:
                                if not rule_exist:
                                    d['location'].append(location)

                                    # Check if score_logbook already has entry for today:
                                    today_found = False
                                    for logbook_entry in d['score_logbook']:
                                        if logbook_entry['date'] == date_today:
                                            logbook_entry['comment'] += f'. Detection rule added: {rule_name}'
                                            today_found = True
                                            break

                                    if not today_found:
                                        d['score_logbook'].append({
                                            'date': date_today,
                                            'score': self._get_latest_score(d),
                                            'comment': f'Auto added by Dettectinator. TODO: Check score. Detection rule added: {rule_name}'
                                        })
                                        results.append(f'Check score for technique {technique_id}. Detection rule(s) added for applicable_to {d["applicable_to"]}.')

                                break
                    else:
                        # applicable_to not present, add new detection object:
                        new_detection = deepcopy(YAML_OBJ_DETECTION)
                        new_detection['applicable_to'] = rule_data['applicable_to']
                        new_detection['location'] = [location]
                        new_detection['score_logbook'][0]['date'] = date_today
                        new_detection['score_logbook'][0]['score'] = 1
                        new_detection['score_logbook'][0]['comment'] = f'Auto added by Dettectinator. TODO: Check score. applicable_to with detection rule added: {rule_name}'
                        yaml_technique['detection'].append(new_detection)
                        results.append(f'Check score for technique {technique_id}. applicable_to is new: {str(rule_data["applicable_to"])}.')
                else:
                    # Technique not present in YAML, add:
                    new_technique = deepcopy(YAML_OBJ_TECHNIQUE)
                    new_technique['technique_id'] = technique_id
                    new_technique['technique_name'] = attack_technique['name']
                    new_technique['detection'] = []
                    new_technique['visibility'] = []
                    new_detection = deepcopy(YAML_OBJ_DETECTION)
                    new_detection['applicable_to'] = rule_data['applicable_to']
                    new_detection['location'] = [location]
                    new_detection['score_logbook'][0]['date'] = date_today
                    new_detection['score_logbook'][0]['score'] = 1
                    new_detection['score_logbook'][0]['comment'] = f'Auto added by Dettectinator. TODO: Check score. Technique with detection rule added: {rule_name}'
                    new_technique['detection'].append(new_detection)
                    new_visibility = deepcopy(YAML_OBJ_VISIBILITY)
                    new_technique['visibility'].append(new_visibility)
                    self._yaml_content['techniques'].append(new_technique)
                    results.append(f'Check score for technique {technique_id}. Technique is new.')

        return warnings, results

    def _delete_rules(self, detection_rules: dict, check_unused_detections: bool, clean_unused_detections: bool, location_prefix_unused_detections: str, check_unused_applicable_to: bool, clean_unused_applicable_to: bool, date_today: datetime) -> list:
        """
        Removes detection rules from the techniques YAML file which are not in the given list of detection_rules. A score_logbook entry is added for every change.
        """
        # First categorize all detection rules per applicable_to:
        warnings, results = [], []
        applicable_to_rules = {}
        for rule_name, rule_data in detection_rules.items():
            if str(rule_data['applicable_to']) not in applicable_to_rules.keys():
                applicable_to_rules[str(rule_data['applicable_to'])] = {}  # cast key to string because a list cannot be a dictionary index name
            applicable_to_rules[str(rule_data['applicable_to'])][rule_name] = rule_data

        # Loop through all techniques in techniques YAML file:
        for yaml_technique in self._yaml_content['techniques']:
            # Look through all detection objects:
            to_remove = []
            for detection in yaml_technique['detection']:
                # Proceed if applicable to from YAML file exists in detection rules list:
                if str(detection['applicable_to']) in applicable_to_rules.keys():
                    # Get the rules for this applicable_to with the same ATT&CK technique:
                    rules_for_app_to = []
                    for rule_name, rule_data in applicable_to_rules[str(detection['applicable_to'])].items():
                        if yaml_technique['technique_id'] in rule_data['techniques']:
                            location = rule_name if rule_data['location_prefix'] == '' else rule_data['location_prefix'] + ': ' + rule_name
                            rules_for_app_to.append(location)

                    # Make a copy to loop through while editting the original location field:
                    locations = deepcopy(detection['location'])
                    for loc in locations:
                        if loc.startswith(location_prefix_unused_detections) and loc not in rules_for_app_to:
                            if check_unused_detections and not clean_unused_detections:
                                warnings.append('Rule from YAML not found in rules list: ' + loc)
                            elif check_unused_detections and clean_unused_detections:
                                detection['location'].remove(loc)
                                score = 0 if len(detection['location']) == 0 else self._get_latest_score(detection)

                                # Check if score_logbook already has entry for today:
                                today_found = False
                                for logbook_entry in detection['score_logbook']:
                                    if logbook_entry['date'] == date_today:
                                        logbook_entry['score'] = score
                                        logbook_entry['comment'] += f'. Detection rule removed: {loc}'
                                        today_found = True
                                        break

                                if not today_found:
                                    detection['score_logbook'].append({
                                        'date': date_today,
                                        'score': score,
                                        'comment': f'Auto added by Dettectinator. TODO: Check score. Detection rule removed: {loc}'
                                    })
                                    results.append(f'Check score for technique {yaml_technique["technique_id"]}. Detection rule(s) removed.')
                else:
                    if check_unused_applicable_to and not clean_unused_applicable_to:
                        warnings.append(f'YAML applicable_to for technique "{yaml_technique["technique_id"]}" not in rules list: {str(detection["applicable_to"])}.')
                    elif check_unused_applicable_to and clean_unused_applicable_to:
                        to_remove.append(detection)

            for item in to_remove:
                warnings.append(f'YAML applicable_to for technique "{yaml_technique["technique_id"]}" not in rules list, so removed: {str(detection["applicable_to"])}.')
                yaml_technique['detection'].remove(item)

        return warnings, results

    def _load_techniques(self) -> None:
        """
        Loads the normalized techniques (including detection and visibility properties) from the techniques YAML file.
        """
        my_techniques = {}

        for d in self._yaml_content['techniques']:
            if 'detection' in d:
                # Add detection items:
                if isinstance(d['detection'], dict):  # There is just one detection entry
                    d['detection'] = self._set_yaml_dv_comments(d['detection'])
                    self._add_entry_to_list_in_dictionary(my_techniques, d['technique_id'], 'detection', d['detection'])
                elif isinstance(d['detection'], list):  # There are multiple detection entries
                    for de in d['detection']:
                        de = self._set_yaml_dv_comments(de)
                        self._add_entry_to_list_in_dictionary(my_techniques, d['technique_id'], 'detection', de)

            if 'visibility' in d:
                # Add visibility items
                if isinstance(d['visibility'], dict):  # There is just one visibility entry
                    d['visibility'] = self._set_yaml_dv_comments(d['visibility'])
                    self._add_entry_to_list_in_dictionary(my_techniques, d['technique_id'], 'visibility', d['visibility'])
                elif isinstance(d['visibility'], list):  # There are multiple visibility entries
                    for de in d['visibility']:
                        de = self._set_yaml_dv_comments(de)
                        self._add_entry_to_list_in_dictionary(my_techniques, d['technique_id'], 'visibility', de)

        self.techniques = my_techniques

    def _get_technique(self, technique_id: str) -> object:
        """
        Gets the technique from the normalized YAML techniques.
        """
        for tech_id, tech in self.techniques.items():
            if tech_id == technique_id:
                return tech

    def _get_technique_from_yaml(self, technique_id: str) -> object:
        """
        Gets the technique from the techniques YAML file.
        """
        for tech in self._yaml_content['techniques']:
            if tech['technique_id'] == technique_id:
                return tech

    @staticmethod
    def _get_latest_score(yaml_object: object) -> int or None:
        """
        Return the latest score present in the score_logbook
        :param yaml_object: a detection or visibility YAML object
        :return: score as an integer or None
        """
        score_obj = DettectTechniquesAdministration._get_latest_score_obj(yaml_object)
        if score_obj:
            return score_obj['score']
        else:
            return None

    @staticmethod
    def _get_latest_score_obj(yaml_object: object) -> object:
        """
        Get the score object in the score_logbook by date
        :param yaml_object: a detection or visibility YAML object
        :return: the latest score object
        """
        if not isinstance(yaml_object['score_logbook'], list):
            yaml_object['score_logbook'] = [yaml_object['score_logbook']]

        if len(yaml_object['score_logbook']) > 0 and 'date' in yaml_object['score_logbook'][0]:
            # for some weird reason 'sorted()' provides inconsistent results
            newest_score_obj = None
            newest_date = None
            for score_obj in yaml_object['score_logbook']:
                score_obj_date = score_obj['date']

                if not newest_score_obj or (score_obj_date and score_obj_date > newest_date):
                    newest_date = score_obj_date
                    newest_score_obj = score_obj

            return newest_score_obj
        else:
            return None


class DettectDataSourcesAdministration(DettectBase):
    """
    Create or modify a DeTT&CT data source administration YAML file.
    """

    def __init__(self, filename: str = None, domain: str = None, local_stix_path: str = None) -> None:
        super(DettectDataSourcesAdministration, self).__init__(filename, domain, local_stix_path)

        self._system_applicable_to_values = ['all']
        for s in self._yaml_content['systems']:
            self._system_applicable_to_values.append(s['applicable_to'])

        self._get_data_components_from_cti()

    def update_data_sources(self, data_sources: dict, check_unused_data_sources: bool = False, clean_unused_data_sources: bool = False) -> list:
        """
        Updates the data source YAML file with the given data sources. All data quality scores are set to "1" for new data sources.

        :param  data_sources: a dictionary of dictionaries containing data source name as key with the following data in a list: applicable_to,
                              products, available_for_data_analytics and data_quality. data_quality may be omitted of remain empty. In that case
                              default data quality scores of "1" will be used.
                {
                    { 'Process Creation': [{'applicable_to': ['all'],
                                           'products': ['Windows Event Log'],
                                           'available_for_data_analytics': True,
                                           'data_quality': {'device_completeness': 1,
                                                            'data_field_completeness': 1,
                                                            'timeliness': 1,
                                                            'consistency': 1,
                                                            'retention': 1}
                                           }]
                    }
                }
        :param  check_unused_data_sources: boolean. When True, all data sources in YAML file will be checked against given data_sources list.
        :param  clean_unused_data_sources: boolean. When True, all data sources in YAML file that don't exist in the given data_sources list will be removed.
        :return a list with results containing warnings or errors during the update process.
        """
        date_today = datetime.utcnow().replace(hour=0, minute=0, second=0, microsecond=0)

        warnings, results = self._add_data_sources(data_sources, date_today)

        if check_unused_data_sources:
            w, r = self._delete_data_sources(data_sources, clean_unused_data_sources)
            warnings += w
            results += r

        return warnings, results

    def _add_data_sources(self, data_sources: dict, date_today: datetime) -> list:
        """
        Adds new data sources to the data source YAML file. A comment is added for every change.
        """
        warnings, results = [], []

        # Loop through all given data sources:
        for data_source_name, data_source_applicable_to in data_sources.items():
            if data_source_name.lower() not in self.data_components.keys():
                results.append(f'Data source "{data_source_name}" does not exist in ATT&CK ({self.domain}). Skipping.')
                continue
            else:
                data_source_name = self.data_components[data_source_name.lower()]

            # Check if data source is already present in YAML file:
            yaml_data_source = self._get_data_source_from_yaml(data_source_name)
            if yaml_data_source is None:
                # Data source not present in YAML, add:
                new_data_source_object = deepcopy(YAML_OBJ_DATA_SOURCES)
                new_data_source_object['data_source_name'] = data_source_name
                new_data_source_object['data_source'] = []
                self._yaml_content['data_sources'].append(new_data_source_object)
                yaml_data_source = new_data_source_object
                results.append(f'Check data quality scores for data source {data_source_name}. Data source is new.')

            # Loop through all given applicable_to values:
            for data_source_data in data_source_applicable_to:
                # Health check: is given applicable_to present in systems object. If not: it will be added with platform=all
                for a in data_source_data['applicable_to']:
                    if a not in self._system_applicable_to_values:
                        self._yaml_content['systems'].append({'applicable_to': a, 'platform': ['all']})
                        self._system_applicable_to_values.append(a)
                        results.append(f'Applicable_to value "{str(data_source_data["applicable_to"])}" of data source "{data_source_name}" not in systems object of data source YAML file. Added to systems object with platform=all. Please review this new entry.')

                # Check if applicable_to is already there:
                applicable_to_list = [d['applicable_to'] for d in yaml_data_source['data_source']]
                if data_source_data['applicable_to'] not in applicable_to_list:
                    # applicable_to not present, add new data source object:
                    new_data_source = deepcopy(YAML_OBJ_DATA_SOURCE)
                    new_data_source['applicable_to'] = data_source_data['applicable_to']
                    new_data_source['date_registered'] = date_today
                    new_data_source['date_connected'] = date_today
                    new_data_source['products'] = data_source_data['products']
                    new_data_source['available_for_data_analytics'] = data_source_data['available_for_data_analytics']
                    default_scores, scores_changed = self._set_data_quality(new_data_source, data_source_data)
                    new_data_source['comment'] = 'Auto added by Dettectinator.' + (' TODO: Check data quality scores, default values used.' if default_scores else '')
                    yaml_data_source['data_source'].append(new_data_source)
                    results.append(f'Check data quality scores for data source {data_source_name}. applicable_to is new: {str(data_source_data["applicable_to"])}')
                else:
                    # applicable_to present, get the object and update:
                    for data_source in yaml_data_source['data_source']:
                        if data_source['applicable_to'] == data_source_data['applicable_to']:
                            changed = False
                            if data_source['products'] != data_source_data['products']:
                                data_source['products'] = data_source_data['products']
                                changed = True
                            if data_source['available_for_data_analytics'] != data_source_data['available_for_data_analytics']:
                                data_source['available_for_data_analytics'] = data_source_data['available_for_data_analytics']
                                changed = True
                            default_scores, scores_changed = self._set_data_quality(data_source, data_source_data)
                            if scores_changed:
                                changed = True
                            if changed:
                                data_source['comment'] = 'Auto updated by Dettectinator. TODO: Check data quality scores.'
                                results.append(f'Check data quality scores for data source {data_source_name}. Data source with applicable_to {str(data_source_data["applicable_to"])} is updated.')
                            break

        return warnings, results

    def _delete_data_sources(self, data_sources: list, clean_unused_data_sources: bool) -> list:
        """
        Removes data sources from the data source YAML file which are not in the given data_sources list. A comment is added for every change.
        """
        warnings, results = [], []

        data_sources_lowercase = [ds.lower() for ds in data_sources.keys()]

        # Loop through all data sources in YAML file:
        to_remove = []
        for yaml_data_source in self._yaml_content['data_sources']:
            # Check if data source is present in data_sources list:
            if yaml_data_source['data_source_name'].lower() in data_sources_lowercase:
                # Data source is present, now check if applicable_to's from YAML are present in data_sources list.

                # Assemble all applicable_to's from data_source list:
                applicable_to_ds = []
                for data_source_applicable_to in data_sources.values():
                    for data_source_data in data_source_applicable_to:
                        applicable_to_ds.append(str(data_source_data['applicable_to']))

                # Make a copy to loop through applicable_to's in YAML (so that original can be mutated):
                applicable_to_yaml = deepcopy(yaml_data_source['data_source'])
                for app_to in applicable_to_yaml:
                    if str(app_to['applicable_to']) not in applicable_to_ds:
                        yaml_data_source['data_source'].remove(app_to)
                        warnings.append(f'YAML applicable_to value "{str(app_to["applicable_to"])}" of data source "{yaml_data_source["data_source_name"]}" not in given data sources list, so removed.')
            else:
                # Data source in YAML but not in data_sources list. So remove.
                if not clean_unused_data_sources:
                    warnings.append(f'YAML data source "{yaml_data_source["data_source_name"]}" not in given data sources list.')
                else:
                    to_remove.append(yaml_data_source)

        for item in to_remove:
            self._yaml_content['data_sources'].remove(item)
            warnings.append(f'YAML data source "{item["data_source_name"]}" not in given data sources list, so removed.')

        return warnings, results

    @staticmethod
    def _set_data_quality(new_data_source_object, data_source_data):
        if 'data_quality' in data_source_data.keys() and data_source_data['data_quality'] != {}:
            changed = False
            if new_data_source_object['data_quality']['device_completeness'] != data_source_data['data_quality']['device_completeness']:
                new_data_source_object['data_quality']['device_completeness'] = data_source_data['data_quality']['device_completeness']
                changed = True
            if new_data_source_object['data_quality']['data_field_completeness'] != data_source_data['data_quality']['data_field_completeness']:
                new_data_source_object['data_quality']['data_field_completeness'] = data_source_data['data_quality']['data_field_completeness']
                changed = True
            if new_data_source_object['data_quality']['timeliness'] != data_source_data['data_quality']['timeliness']:
                new_data_source_object['data_quality']['timeliness'] = data_source_data['data_quality']['timeliness']
                changed = True
            if new_data_source_object['data_quality']['consistency'] != data_source_data['data_quality']['consistency']:
                new_data_source_object['data_quality']['consistency'] = data_source_data['data_quality']['consistency']
                changed = True
            if new_data_source_object['data_quality']['retention'] != data_source_data['data_quality']['retention']:
                new_data_source_object['data_quality']['retention'] = data_source_data['data_quality']['retention']
                changed = True
            return False, changed
        else:
            new_data_source_object['data_quality']['device_completeness'] = 1
            new_data_source_object['data_quality']['data_field_completeness'] = 1
            new_data_source_object['data_quality']['timeliness'] = 1
            new_data_source_object['data_quality']['consistency'] = 1
            new_data_source_object['data_quality']['retention'] = 1
            return True, False

    def _get_data_source_from_yaml(self, data_source_name: str) -> object:
        """
        Gets the data source from the data source YAML file.
        """
        for ds in self._yaml_content['data_sources']:
            if ds['data_source_name'] == data_source_name:
                return ds

    def _get_data_components_from_cti(self) -> object:
        """
        Get all data component STIX objects from CTI for the provided ATT&CK Matrix
        :param matrix: ATT&CK Matrix
        :return: list of data component STIX objects
        """
        if self.domain == 'enterprise-attack':
            data_components = self.mitre.TC_ENTERPRISE_SOURCE.query(Filter("type", "=", "x-mitre-data-component"))
        elif self.domain == 'ics-attack':
            # ICS data components are not yet in CTI, so this will not work
            data_components = self.mitre.TC_ICS_SOURCE.query(Filter("type", "=", "x-mitre-data-component"))
        elif self.domain == 'mobile-attack':
            # Mobile data components are not yet in CTI, so this will not work
            data_components = self.mitre.TC_MOBILE_SOURCE.query(Filter("type", "=", "x-mitre-data-component"))

        self.data_components = {}
        for data_component in data_components:
            self.data_components[data_component['name'].lower()] = data_component['name']


if __name__ == '__main__':
    # Where being run from the command line here, start CLI logic
    cli_mod = importlib.import_module('cli')
    cli= getattr(cli_mod, 'CommandLine')()
    cli.start()



"""
Dettectinator - The Python library to your DeTT&CT YAML files.
Authors:
    Martijn Veken, Sirius Security
    Ruben Bouman, Sirius Security
License: GPL-3.0 License
"""

from plugins.support.msal_patch import PublicClientApplicationPatch
import msal
import datetime
import json
import requests


class Azure:
    """
    Class for authenticating agaings Azure AD
    """

    def __init__(self):
        pass

    @staticmethod
    def connect_device_flow(app_id: str, tenant_id: str, endpoint: str) -> str:
        """
        Login to Azure AD using  Device Flow authentication
        :return: Access token to use with the API
        """
        authority = 'https://login.microsoftonline.com/' + tenant_id
        scope = [endpoint + '/.default']

        app = PublicClientApplicationPatch(app_id, authority=authority)

        # Insert a User-Agent header to mimic a Windows device
        # This can be changed to adapt to certain Conditional Access Policies
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36 Edg/106.0.1370.34'  # This is another valid field
        }

        flow = app.initiate_device_flow(scopes=scope, headers=headers)
        if 'user_code' not in flow:
            raise Exception('Azure: Failed to create device flow. Err: %s' % json.dumps(flow, indent=4))

        print(flow['message'])
        print('Waiting for authentication...\n')
        logon_result = app.acquire_token_by_device_flow(flow, headers=headers)

        if 'access_token' in logon_result:
            print('You have been succesfully logged in: ')
            print(f'Name: {logon_result["id_token_claims"]["name"]}')
            print(f'UPN: {logon_result["id_token_claims"]["preferred_username"]}')
            print(f'Token expiration: {datetime.datetime.fromtimestamp(logon_result["id_token_claims"]["exp"]).isoformat()}')
            return logon_result['access_token']
        else:
            raise Exception('Azure: Failed to logon to Azure AD.')

    @staticmethod
    def connect_client_secret(app_id: str, tenant_id: str, endpoint: str, secret: str) -> str:
        """
        Login to Azure AD using Client secret authentication
        :return: Access token to use with the API
        """
        authority = 'https://login.microsoftonline.com/' + tenant_id
        scope = [endpoint + '/.default']

        app = msal.ConfidentialClientApplication(app_id, authority=authority, client_credential=secret)

        logon_result = app.acquire_token_for_client(scopes=scope)

        if "access_token" in logon_result:
            print('You have been succesfully logged in.')
            print(f'Token expires in {logon_result["expires_in"]} seconds.')
            return logon_result['access_token']
        else:
            raise Exception('Azure: Failed to logon to Azure AD.')


class Tanium:
    """
    Class to authenticate against Tanium
    """

    def __init__(self):
        pass

    @staticmethod
    def connect_http(user: str, password: str, login_url: str) -> str:
        """
        Logs in to the Tanium host and saves the session ticket.
        """
        data = {'username': user, 'password': password}
        r = requests.post(login_url, data=json.dumps(data), verify=False)
        if r.status_code == 200:
            return r.json()['data']['session']
        else:
            raise Exception('Tanium: login failed.')



"""
The contents of this file are taken with permission from the DeTT&CT project (http://github.com/rabobank-cdc/DeTTECT).
"""

PLATFORMS_ENTERPRISE = {'pre': 'PRE', 'windows': 'Windows', 'macos': 'macOS', 'linux': 'Linux', 'office 365': 'Office 365',
                        'azure ad': 'Azure AD', 'google workspace': 'Google Workspace', 'iaas': 'IaaS', 'saas': 'SaaS',
                        'network': 'Network', 'containers': 'Containers'}

PLATFORMS_ICS = {'control server': 'Control Server', 'data historian': 'Data Historian', 'device configuration/parameters': 'Device Configuration/Parameters',
                 'engineering workstation': 'Engineering Workstation', 'field controller/rtu/plc/ied': 'Field Controller/RTU/PLC/IED',
                 'human-machine interface': 'Human-Machine Interface', 'input/output server': 'Input/Output Server',
                 'safety instrumented system/protection relay': 'Safety Instrumented System/Protection Relay', 'windows': 'Windows',
                 'none': 'None'}

PLATFORMS_MOBILE = {'android': 'Android', 'ios': 'iOS'}

YAML_OBJ_DETECTION = {'applicable_to': ['all'],
                      'location': [''],
                      'comment': '',
                      'score_logbook':
                          [
                              {'date': None,
                               'score': -1,
                               'comment': ''}
                      ]}

YAML_OBJ_VISIBILITY = {'applicable_to': ['all'],
                       'comment': '',
                       'score_logbook':
                           [
                               {'date': None,
                                'score': 0,
                                'comment': '',
                                'auto_generated': True}
                           ]
                       }

YAML_OBJ_TECHNIQUE = {'technique_id': '',
                      'technique_name': '',
                      'detection': [YAML_OBJ_DETECTION],
                      'visibility': [YAML_OBJ_VISIBILITY]}

YAML_OBJ_NEW_TECHNIQUES_FILE = {'version': 1.2,
                                'file_type': 'technique-administration',
                                'name': 'new',
                                'domain': 'enterprise-attack',
                                'platform': ['all'],
                                'techniques': []
                               }

YAML_OBJ_DATA_SOURCE = {'applicable_to': ['all'],
                        'date_registered': None,
                        'date_connected': None,
                        'products': [],
                        'available_for_data_analytics': False,
                        'comment': '',
                        'data_quality': {
                            'device_completeness': 0,
                            'data_field_completeness': 0,
                            'timeliness': 0,
                            'consistency': 0,
                            'retention': 0}
                        }

YAML_OBJ_DATA_SOURCES = {'data_source_name': '',
                         'data_source': [YAML_OBJ_DATA_SOURCE]}

YAML_OBJ_NEW_DATA_SOURCES_FILE = {'version': 1.1,
                                  'file_type': 'data-source-administration',
                                  'name': 'new',
                                  'domain': 'enterprise-attack',
                                  'systems': [{ 'applicable_to': 'default', 'platform': ['all'] }],
                                  'data_sources': []
                                 }

YAML_OBJ_NEW_GROUPS_FILE = {'version': 1.0,
                            'file_type': 'group-administration',
                            'domain': 'enterprise-attack',
                            'platform': ['all'],
                            'groups': []
                           }


"""
Dettectinator - The Python library to your DeTT&CT YAML files.
Authors:
    Martijn Veken, Sirius Security
    Ruben Bouman, Sirius Security
License: GPL-3.0 License

Parts this module have been taken from the Microsoft MSAL library for Python:
https://github.com/AzureAD/microsoft-authentication-library-for-python/

On those parts the following terms and conditions are applicable:

The MIT License (MIT)

Copyright (c) Microsoft Corporation.
All rights reserved.

This code is licensed under the MIT License.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files(the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions :

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from msal import PublicClientApplication
from msal.application import _clean_up, _merge_claims_challenge_and_capabilities
import msal


class PublicClientApplicationPatch(PublicClientApplication):
    """
    This class patches the msal.PublicClientApplication class so that it accepts additional headers for the
    device flow initiation request. This allows specifying the OS/Platform type by setting an appropriate
    user agent value.
    """

    def __init__(self, client_id, client_credential=None, **kwargs):
        super().__init__(client_id, client_credential, **kwargs)

    def initiate_device_flow(self, scopes=None, **kwargs):
        """Initiate a Device Flow instance,
        which will be used in :func:`~acquire_token_by_device_flow`.

        :param list[str] scopes:
            Scopes requested to access a protected API (a resource).
        :return: A dict representing a newly created Device Flow object.

            - A successful response would contain "user_code" key, among others
            - an error response would contain some other readable key/value pairs.
        """
        correlation_id = msal.telemetry._get_new_correlation_id()

        # PATCH
        # Add the supplied headers to the client request id header
        headers = kwargs.pop("headers", {})
        headers.update({msal.telemetry.CLIENT_REQUEST_ID: correlation_id})
        # /PATCH

        flow = self.client.initiate_device_flow(
            scope=self._decorate_scope(scopes or []),
            headers=headers,
            **kwargs)
        flow[self.DEVICE_FLOW_CORRELATION_ID] = correlation_id
        return flow

    def acquire_token_by_device_flow(self, flow, claims_challenge=None, **kwargs):
        """Obtain token by a device flow object, with customizable polling effect.

        :param dict flow:
            A dict previously generated by :func:`~initiate_device_flow`.
            By default, this method's polling effect  will block current thread.
            You can abort the polling loop at any time,
            by changing the value of the flow's "expires_at" key to 0.
        :param claims_challenge:
            The claims_challenge parameter requests specific claims requested by the resource provider
            in the form of a claims_challenge directive in the www-authenticate header to be
            returned from the UserInfo Endpoint and/or in the ID Token and/or Access Token.
            It is a string of a JSON object which contains lists of claims being requested from these locations.

        :return: A dict representing the json response from AAD:

            - A successful response would contain "access_token" key,
            - an error response would contain "error" and usually "error_description".
        """
        telemetry_context = self._build_telemetry_context(
            self.ACQUIRE_TOKEN_BY_DEVICE_FLOW_ID,
            correlation_id=flow.get(self.DEVICE_FLOW_CORRELATION_ID))

        # PATCH
        # Add the supplied headers to the client request id header
        headers = kwargs.pop("headers", {})
        headers.update(telemetry_context.generate_headers())
        # /PATCH

        response = _clean_up(self.client.obtain_token_by_device_flow(
            flow,
            data=dict(
                kwargs.pop("data", {}),
                code=flow["device_code"],  # 2018-10-4 Hack:
                    # during transition period,
                    # service seemingly need both device_code and code parameter.
                claims=_merge_claims_challenge_and_capabilities(
                    self._client_capabilities, claims_challenge),
                ),
            headers=headers,
            **kwargs))
        telemetry_context.update_telemetry(response)
        return response



import setuptools

long_description = """# Dettectinator
Dettectinator - The Python library to your [DeTT&amp;CT](https://github.com/rabobank-cdc/DeTTECT) YAML files.

Dettectinator is built to be included in your SOC automation tooling. It can be included as a Python library or it can be used via the command line.

Dettectinator provides plugins to read detections from your SIEM or EDR and create/update the DeTT&CT YAML file, so that you can use it to visualize your ATT&CK detection coverage in the ATT&CK Navigator.

Currently de CLI is limited to processing detections through these plugins, the library can also be used for processing data sources.

See the [documentation](https://github.com/siriussecurity/dettectinator) for more information on how to use it."""

setuptools.setup(
    name="dettectinator",
    version="1.1.0",
    author="Sirius Security",
    description="Dettectinator - The Python library to your DeTT&CT YAML files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/siriussecurity/dettectinator",
    project_urls={
        "Documentation": "https://github.com/siriussecurity/dettectinator/wiki",
        "Code": "https://github.com/siriussecurity/dettectinator",
        "Issue tracker": "https://github.com/siriussecurity/dettectinator/issues",
    },
    keywords="mitre attack dettect soc threat hunting",
    packages=setuptools.find_packages(exclude=["examples", "tests"]),
    install_requires=["requests", "ruamel.yaml", "attackcti", "python-dateutil", "msal", "stix2", "openpyxl", "suricataparser", "addonfactory-splunk-conf-parser-lib", "pandas", "anyascii"],
    license='GPL3',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Topic :: Security',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.9'
    ],
)


"""
Dettectinator - The Python library to your DeTT&CT YAML files.
Authors:
    Martijn Veken, Sirius Security
    Ruben Bouman, Sirius Security
License: GPL-3.0 License
"""

import json
import xml.etree.ElementTree as ElementTree
import requests
import re
import sys
import os

from argparse import ArgumentParser
from collections.abc import Iterable
from xml.etree.ElementTree import Element

try:
    # When dettectinator is installed as python library
    from dettectinator.plugins.support.authentication import Azure
except ModuleNotFoundError:
    # When dettectinator is not installed as python library
    sys.path.append(os.path.dirname(os.path.abspath(__file__).replace('plugins', '')))
    from plugins.support.authentication import Azure


class DatasourceBase:
    """
    Base class for importing datasource/product data
    """

    def __init__(self, parameters: dict) -> None:
        self._parameters = parameters

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        raise NotImplementedError()

    def get_attack_datasources(self, default_applicable_to: list) -> dict:
        """
        Retrieves datasource/product data from a data source
        :param default_applicable_to: Systems that the data sources are applicable to.
        :return: Dictionary, example: {"User Account Creation":[{"applicable_to":["test"],"available_for_data_analytics":true,"products":["DeviceEvents: UserAccountCreated"]}]}
        """

        data_sources = {}

        for datasource, product, applicable_to in self.get_data_from_source():
            applicable_to = applicable_to or default_applicable_to

            if datasource not in data_sources.keys():
                record = {'applicable_to': applicable_to, 'available_for_data_analytics': True, 'products': []}
                data_sources[datasource] = [record]
            else:
                record = data_sources[datasource][0]

            if product not in record['products']:
                record['products'].append(product)

        return data_sources

    def get_data_from_source(self) -> Iterable:
        """
        Gets the datasource/product data from the source.
        :return: Iterable, yields datasource, product, applicable_to
        """
        raise NotImplementedError()


class DatasourceCsv(DatasourceBase):
    """
    Import data from a CSV file, formatted Datasource,Product
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)
        if 'file' not in self._parameters:
            raise Exception('DatasourceCsv: "file" parameter is required.')

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        parser.add_argument('--file', help='Path of the csv file to import', required=True)

    def get_data_from_source(self) -> Iterable:
        """
        Gets the datasource/product data from the source.
        :return: Iterable, yields datasource, product, applicable_to
        """
        file = self._parameters['file']
        print(f'Reading data from "{file}"')

        with open(file) as f:
            lines = f.readlines()

        for detection in lines:
            parts = detection.split(',')
            data_source = parts[0].strip()
            product = parts[1].strip()
            yield data_source, product, None


class DatasourceExcel(DatasourceBase):
    """
    Import data sources from an Excel file, having a worksheet with two columns: Datasource and Product
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)
        if 'file' not in self._parameters:
            raise Exception('DatasourceExcel: "file" parameter is required.')

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        parser.add_argument('--file', help='Path of the Excel file to import', required=True)

    def get_data_from_source(self) -> Iterable:
        """
        Gets the datasource/product data from the source.
        :return: Iterable, yields datasource, product, applicable_to
        """
        file = self._parameters['file']
        print(f'Reading data from "{file}"')

        import openpyxl
        wb = openpyxl.load_workbook(filename=file, data_only=True)
        sheet = wb.worksheets[0]

        for rowNumber in range(2, sheet.max_row + 1):
            data_source = sheet.cell(row=rowNumber, column=1).value
            product = sheet.cell(row=rowNumber, column=2).value
            yield data_source, product, None


class DatasourceOssemBase(DatasourceBase):
    """
    Base class for importing datasource/product data that is based on OSSEM data
    For information about OSSEM see: https://github.com/OTRF/OSSEM-DM
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)
        self._log_source = None

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        raise NotImplementedError()

    def get_data_from_source(self) -> Iterable:
        """
        Gets the datasource/product data from the source.
        :return: Iterable, yields datasource, product, applicable_to
        """
        raise NotImplementedError()

    def _get_ossem_data(self):
        """
        Retrieves data from the OSSEM ATT&CK mapping
        """
        import pandas

        url = 'https://raw.githubusercontent.com/OTRF/OSSEM-DM/main/use-cases/mitre_attack/attack_events_mapping.csv'
        data = pandas.read_csv(url)
        data.where(data['Log Source'] == self._log_source, inplace=True)
        data.dropna(how='all', inplace=True)
        select = data[['Data Source', 'Component', 'EventID', 'Event Name', 'Filter in Log', 'Audit Category']]
        dict_result = select.to_dict(orient="records")
        return dict_result


class DatasourceDefenderEndpoints(DatasourceOssemBase):
    """
    Class for importing data source information for Microsoft Defender for Endpoints tables.
    Uses OSSEM to generate the overview.
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)
        self._log_source = 'Microsoft Defender for Endpoint'

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        pass

    def get_data_from_source(self) -> Iterable:
        """
        Gets the datasource/product data from the source.
        :return: Iterable, yields datasource, product, applicable_to
        """
        ossem_data = self._get_ossem_data()

        for record in ossem_data:
            action_types = json.loads(record['Filter in Log'].replace('\'', '\"'))
            if len(action_types) > 0:
                for action_type in action_types:
                    yield str(record['Component']).title(), f'{record ["Event Name"]}: {action_type["ActionType"]}', None
            else:
                yield str(record['Component']).title(), record['Event Name'], None


class DatasourceWindowsSysmon(DatasourceOssemBase):
    """
    Class for importing data source information for Sysmon.
    Uses your Sysmon config file and OSSEM to generate the overview.
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)

        if 'sysmon_config' not in self._parameters:
            raise Exception('DatasourceWindowsSysmon: "sysmon_config" parameter is required.')

        self._sysmon_config = parameters['sysmon_config']
        self._log_source = 'Microsoft-Windows-Sysmon'

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        parser.add_argument('--sysmon_config', help='Path of the Sysmon config file.', required=True)

    def get_data_from_source(self) -> Iterable:
        """
        Gets the datasource/product data from the source.
        :return: Iterable, yields datasource, product, applicable_to
        """
        ossem_data = self._get_ossem_data()
        sysmon_config = self._get_sysmon_config()

        for record in ossem_data:
            config_items = sysmon_config.findall(f'.//{record["Audit Category"]}')

            # If this is an event type with an onmatch == include attribute without child items this means
            # that nothing is being logged for this event type
            for config_item in config_items:
                if config_item.attrib['onmatch'] == "include" and len(config_item.getchildren()) == 0:
                    continue

            data_source = str(record['Component']).title()
            product = f'{record["EventID"]}: {record["Event Name"]}'
            yield data_source, product, None

    def _get_sysmon_config(self) -> Element:
        """
        Gets the Sysmon config from the filesystem
        """
        tree = ElementTree.parse(self._sysmon_config)
        root = tree.getroot()
        return root


class DatasourceWindowsSecurityAuditing(DatasourceOssemBase):
    """
    Class for importing data source information for Windows Security Auditing event logging.
    Uses the event id's logged in the last 30 days and OSSEM to generate the overview.
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)

        if 'app_id' not in self._parameters:
            raise Exception(f'DatasourceWindowsSecurityAuditing: "app_id" parameter is required.')

        if 'tenant_id' not in self._parameters:
            raise Exception(f'DatasourceWindowsSecurityAuditing: "tenant_id" parameter is required.')

        if 'workspace_id' not in self._parameters:
            raise Exception(f'DatasourceWindowsSecurityAuditing: "workspace_id" parameter is required.')

        self._app_id = self._parameters['app_id']
        self._tenant_id = self._parameters['tenant_id']
        self._workspace_id = self._parameters['workspace_id']
        self._secret = self._parameters.get('secret', None)
        self._endpoint = 'https://api.loganalytics.io'

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        parser.add_argument('--app_id', help='Azure application id', required=True)
        parser.add_argument('--tenant_id', help='Azure tenant id', required=True)
        parser.add_argument('--workspace_id', help='Azure Sentinel workspace id', required=True)
        parser.add_argument('--secret', help='Azure client secret')

    def _connect_to_azure(self, endpoint: str) -> str:
        if self._secret:
            return Azure.connect_client_secret(self._app_id, self._tenant_id, endpoint, self._secret)
        else:
            return Azure.connect_device_flow(self._app_id, self._tenant_id, endpoint)

    def get_data_from_source(self) -> Iterable:
        """
        Gets the datasource/product data from the source.
        :return: Iterable, yields datasource, product, applicable_to
        """
        access_token = self._connect_to_azure(self._endpoint)
        sentinel_data = self._get_sentinel_data(access_token)

        for record in sentinel_data:
            data_source = record[0].title()
            product = record[1]
            yield data_source, product, None

    def _get_sentinel_data(self, access_token: str) -> list:
        """
        Execute a query on Advanced Hunting to retrieve the datasource/product data
        :param access_token: JWT token to execute the request on the backend
        :return: Dictionary containing the results
        """
        url = f'{self._endpoint}/v1/workspaces/{self._workspace_id}/query'

        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': 'Bearer ' + access_token
        }

        query = '''
        let mapping_table = externaldata(DataSource:string,Component:string,Source:string,Relationship:string,Target:string, OSSEMID:string, EventID:int,EventName:string,EventPlatform:string, LogSource:string, FilterInLog:string,AuditCategory:string,AuditSubCategory:string, Channel:string, EnableCommands:string,GPOAuditPolicy:string)
        [
            h@"https://raw.githubusercontent.com/OTRF/OSSEM-DM/main/use-cases/mitre_attack/attack_events_mapping.csv"
        ]
        with(format="csv")
        | where LogSource == "Microsoft-Windows-Security-Auditing";
        // Get the event id's from the Windows Security Log and join this with the mapping table
        SecurityEvent
        | where TimeGenerated >= ago(30d)
        | where EventSourceName == "Microsoft-Windows-Security-Auditing"
        | distinct EventID
        | join kind = inner mapping_table on EventID
        // Get the distinct data sources
        | project  DataSource = Component, Product = strcat("Microsoft-Windows-Security-Auditing: ", tostring(EventID))
        | distinct DataSource, Product
        '''

        data = json.dumps({'query': query}).encode('utf-8')

        response = requests.post(url=url, headers=headers, data=data)

        if response.status_code != requests.codes['ok']:
            # Raise an exception to handle hitting API limits
            if response.status_code == requests.codes['too_many_requests']:
                raise ConnectionRefusedError('DatasourceWindowsSecurityAuditing: You have likely hit the API limit. ')
            response.raise_for_status()

        json_response = response.json()
        result = json_response['tables'][0]['rows']

        return result



"""
Dettectinator - The Python library to your DeTT&CT YAML files.
Authors:
    Martijn Veken, Sirius Security
    Ruben Bouman, Sirius Security
License: GPL-3.0 License
"""

from argparse import ArgumentParser
from collections.abc import Iterable
import json
import os
import sys
import re
import requests
import urllib3

try:
    # When dettectinator is installed as python library
    from dettectinator.plugins.support.authentication import Azure, Tanium
except ModuleNotFoundError:
    # When dettectinator is not installed as python library
    sys.path.append(os.path.dirname(os.path.abspath(__file__).replace('plugins', '')))
    from plugins.support.authentication import Azure, Tanium


# Disable SSL certificate warnings for dev purposes:
urllib3.disable_warnings()


class TechniqueBase:
    """
    Base class for importing use-case/technique data
    """

    def __init__(self, parameters: dict) -> None:
        self._parameters = parameters

        self._re_include = self._parameters.get('re_include', None)
        self._re_exclude = self._parameters.get('re_exclude', None)
        self._location_prefix = self._parameters.get('location_prefix', '')

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        parser.add_argument('-ri', '--re_include', help='Regex for detection names that should be included.',
                            default=None)
        parser.add_argument('-re', '--re_exclude', help='Regex for detection names that should be excluded.',
                            default=None)
        parser.add_argument('-l', '--location_prefix',
                            help='Location of the detection, will be prepended to the detection name.', default='')
        parser.add_argument('-clp', '--clean_unused_location_prefix', action='store_true',
                            help='Clean unused detections based on location_prefix.')

    def get_attack_techniques(self, default_applicable_to: list) -> dict:
        """
        Retrieves use-case/technique data from a data source
        :param default_applicable_to: Default value for systems that the detections are applicable to (can be overrule from source)
        :return: Dictionary, example: {'Detection A': {'applicable_to': ['all'], 'location_prefix': 'SIEM', 'techniques': ['T1055']}}
        """

        use_cases = {}

        for technique, use_case, applicable_to in self.get_data_from_source():
            # Exclude all detections that match the exclude-pattern
            if self._re_exclude and not re.match(self._re_exclude, use_case) is None:
                continue

            # Include all detections that match the include-pattern
            if self._re_include and re.match(self._re_include, use_case) is None:
                continue

            applicable_to = applicable_to or default_applicable_to

            if use_case in use_cases.keys():
                use_cases[use_case]['techniques'].append(technique)
            else:
                use_cases[use_case] = {'applicable_to': applicable_to,
                                       'location_prefix': self._location_prefix,
                                       'techniques': [technique]}

        return use_cases

    def get_data_from_source(self) -> Iterable:
        """
        Gets the use-case/technique data from the source.
        :return: Iterable, yields technique, detection, applicable_to
        """
        raise NotImplementedError()


class TechniqueCsv(TechniqueBase):
    """
    Import data from a CSV file, formatted TechniqueId,UseCase
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)
        if 'file' not in self._parameters:
            raise Exception('TechniqueCsv: "file" parameter is required.')

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        TechniqueBase.set_plugin_params(parser)

        parser.add_argument('--file', help='Path of the csv file to import', required=True)

    def get_data_from_source(self) -> Iterable:
        """
        Gets the use-case/technique data from the source.
        :return: Iterable, yields technique, detection, applicable_to
        """
        file = self._parameters['file']
        print(f'Reading data from "{file}"')

        with open(file) as f:
            lines = f.readlines()

        for detection in lines:
            parts = detection.split(',')
            technique = parts[0].strip()
            use_case = parts[1].strip()
            yield technique, use_case, None


class TechniqueExcel(TechniqueBase):
    """
    Import data from an Excel file, having a worksheet with two columns: TechniqueId and UseCase
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)
        if 'file' not in self._parameters:
            raise Exception('TechniqueExcel: "file" parameter is required.')

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        TechniqueBase.set_plugin_params(parser)

        parser.add_argument('--file', help='Path of the Excel file to import', required=True)

    def get_data_from_source(self) -> Iterable:
        """
        Gets the use-case/technique data from the source.
        :return: Iterable, yields technique, detection, applicable_to
        """
        file = self._parameters['file']
        print(f'Reading data from "{file}"')

        import openpyxl
        wb = openpyxl.load_workbook(filename=file, data_only=True)
        sheet = wb.worksheets[0]

        for rowNumber in range(2, sheet.max_row + 1):
            techniques = sheet.cell(row=rowNumber, column=1).value
            detection = sheet.cell(row=rowNumber, column=2).value
            for technique in techniques.split(','):
                yield technique.strip(), detection, None


class TechniqueAzureAuthBase(TechniqueBase):
    """
    Base class for import plugins that authenticate against Azure AD
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)

        if 'app_id' not in self._parameters:
            raise Exception(f'{self.__class__.__name__}: "app_id" parameter is required.')

        if 'tenant_id' not in self._parameters:
            raise Exception(f'{self.__class__.__name__}: "tenant_id" parameter is required.')

        self._app_id = self._parameters['app_id']
        self._tenant_id = self._parameters['tenant_id']
        self._secret = self._parameters.get('secret', None)

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        TechniqueBase.set_plugin_params(parser)

        parser.add_argument('--app_id', help='Azure application id', required=True)
        parser.add_argument('--tenant_id', help='Azure tenant id', required=True)
        parser.add_argument('--secret', help='Azure client secret')

    def get_data_from_source(self) -> Iterable:
        """
         Gets the use-case/technique data from the source.
         :return: Iterable, yields technique, detection, applicable_to
         """
        raise NotImplementedError()

    def _connect_to_azure(self, endpoint: str) -> str:
        if self._secret:
            return Azure.connect_client_secret(self._app_id, self._tenant_id, endpoint, self._secret)
        else:
            return Azure.connect_device_flow(self._app_id, self._tenant_id, endpoint)


class TechniqueSentinelAlertRules(TechniqueAzureAuthBase):
    """
    Import Analytics Rules from the Sentinel API.
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)

        if 'subscription_id' not in self._parameters:
            raise Exception('TechniqueSentinelAlertRules: "subscription_id" parameter is required.')

        if 'resource_group' not in self._parameters:
            raise Exception('TechniqueSentinelAlertRules: "resource_group" parameter is required.')

        if 'workspace' not in self._parameters:
            raise Exception('TechniqueSentinelAlertRules: "workspace" parameter is required.')

        self._subscription_id = self._parameters['subscription_id']
        self._resource_group = self._parameters['resource_group']
        self._workspace = self._parameters['workspace']
        self._endpoint = 'https://management.azure.com'

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        TechniqueAzureAuthBase.set_plugin_params(parser)

        parser.add_argument('--subscription_id', help='Azure subscription id for Sentinel', required=True)
        parser.add_argument('--resource_group', help='Azure resource group for Sentinel', required=True)
        parser.add_argument('--workspace', help='Azure workspace for Sentinel', required=True)

    def get_data_from_source(self) -> Iterable:
        """
         Gets the use-case/technique data from the source.
         :return: Iterable, yields technique, detection, applicable_to
         """
        access_token = self._connect_to_azure(self._endpoint)
        sentinel_data = self._get_sentinel_data(access_token)

        for record in sentinel_data:
            properties = record['properties']

            if 'techniques' in properties and properties['techniques']:
                for technique in properties['techniques']:
                    use_case = properties['displayName']
                    yield technique, use_case, None

    def _get_sentinel_data(self, access_token: str) -> list:
        """
        Execute a query on Advanced Hunting to retrieve the use-case/technique data
        :param access_token: JWT token to execute the request on the backend
        :return: Dictionary containing the results
        """
        url = f'https://management.azure.com/subscriptions/{self._subscription_id}/resourceGroups/{self._resource_group}/' + \
              f'providers/Microsoft.OperationalInsights/workspaces/{self._workspace}/providers/Microsoft.SecurityInsights/' + \
              'alertRules?api-version=2022-07-01-preview'

        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': 'Bearer ' + access_token
        }
        response = requests.get(url=url, headers=headers)

        if response.status_code != requests.codes['ok']:
            # Raise an exception to handle hitting API limits
            if response.status_code == requests.codes['too_many_requests']:
                raise ConnectionRefusedError('TechniqueSentinelAlertRules: You have likely hit the API limit. ')
            response.raise_for_status()

        json_response = response.json()
        result = json_response['value'] if 'value' in json_response else []

        return result


class TechniqueDefenderIdentityRules(TechniqueBase):
    """
    Import rules for Microsoft Defender for Identity from their Github webpage:
    https://github.com/MicrosoftDocs/ATADocs/tree/master/ATPDocs
    More info:
    https://learn.microsoft.com/en-us/defender-for-identity/alerts-overview
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)

        self.ATP_DOCS = ['https://raw.githubusercontent.com/MicrosoftDocs/ATADocs/master/ATPDocs/compromised-credentials-alerts.md',
                         'https://raw.githubusercontent.com/MicrosoftDocs/ATADocs/master/ATPDocs/domain-dominance-alerts.md',
                         'https://raw.githubusercontent.com/MicrosoftDocs/ATADocs/master/ATPDocs/exfiltration-alerts.md',
                         'https://raw.githubusercontent.com/MicrosoftDocs/ATADocs/master/ATPDocs/lateral-movement-alerts.md',
                         'https://raw.githubusercontent.com/MicrosoftDocs/ATADocs/master/ATPDocs/reconnaissance-alerts.md']

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        TechniqueBase.set_plugin_params(parser)

    def get_data_from_source(self) -> Iterable:
        """
        Gets the use-case/technique data from the source.
        :return: Iterable, yields technique, detection, applicable_to
        """
        for source_url in self.ATP_DOCS:
            resp = requests.get(source_url)
            body = resp.text

            # Remove comments from file, because it may contain commented detection rules.
            while '<!--' in body:
                body = body[0:body.find('<!--')] + body[body.find('-->')+3:]

            regex_title = re.compile('##\s(.*\s\(external\sID\s\d{4}\))')
            regex_tech = re.compile('\((T\d{4})\)')
            regex_subtech = re.compile('(T\d{4}\.\d{3})')

            current_detection = None
            for line in body.splitlines():
                title_match = regex_title.match(line)
                if title_match or current_detection is None:
                    if title_match:
                        current_detection = title_match.group(1)
                        continue
                else:
                    if 'MITRE attack technique' in line and 'N/A' in line:
                        current_detection = None
                    elif 'MITRE attack technique' in line:
                        tech_match = regex_tech.findall(line)
                        if tech_match:
                            for t in tech_match:
                                yield t, current_detection, None
                    elif 'MITRE attack sub-technique' in line and 'N/A' in line:
                        current_detection = None
                    elif 'MITRE attack sub-technique' in line:
                        subtech_match = regex_subtech.findall(line)
                        if subtech_match:
                            for t in subtech_match:
                                yield t, current_detection, None
                            current_detection = None


class TechniqueDefenderAlerts(TechniqueAzureAuthBase):
    """
    Import alerts and techniques from the Microsoft Defender API.
    """
    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)

        self._endpoint = 'https://api.security.microsoft.com'

    def get_data_from_source(self) -> Iterable:
        """
        Gets the use-case/technique data from the source.
        :return: Iterable, yields technique, detection, applicable_to
        """
        access_token = self._connect_to_azure(self._endpoint)
        defender_data = self._get_defender_data(access_token)

        for record in defender_data:
            technique = record['TechniqueId']
            use_case = record['Title'].strip()
            yield technique, use_case, None

    @staticmethod
    def _get_defender_data(access_token: str) -> dict:
        """
        Execute a query on Advanced Hunting to retrieve the use-case/technique data
        :param access_token: JWT token to execute the request on the backend
        :return: Dictionary containing the results
        """
        query = '''
        AlertInfo
        | mv-expand todynamic(AttackTechniques)
        | extend TechniqueId = extract(@'\((T.*)\)', 1, tostring(AttackTechniques))
        | where isnotempty(TechniqueId)
        | distinct TechniqueId, Title
        | order by TechniqueId
        '''

        url = 'https://api.security.microsoft.com/api/advancedhunting/run'
        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': 'Bearer ' + access_token
        }

        data = json.dumps({'Query': query}).encode('utf-8')

        response = requests.post(url=url, headers=headers, data=data)

        if response.status_code != requests.codes['ok']:
            # Raise an exception to handle hitting API limits
            if response.status_code == requests.codes['too_many_requests']:
                raise ConnectionRefusedError('TechniqueDefenderAlerts: You have likely hit the API limit. ')
            response.raise_for_status()

        json_response = response.json()
        result = json_response['Results'] if 'Results' in json_response else {}

        return result


class TechniqueTaniumSignals(TechniqueBase):
    """
    Class for importing signals with ATT&CK technique mapping from Tanium.
    """
    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)

        if 'host' not in self._parameters:
            raise Exception('TechniqueTaniumSignals: "host" parameter is required.')
        if 'user' not in self._parameters:
            raise Exception('TechniqueTaniumSignals: "user" parameter is required.')
        if 'password' not in self._parameters:
            raise Exception('TechniqueTaniumSignals: "password" parameter is required.')
        if 'search_prefix' not in self._parameters:
            raise Exception('TechniqueTaniumSignals: "search_prefix" parameter is required.')

        self._host = self._parameters['host']
        self._user = self._parameters['user']
        self._password = self._parameters['password']
        self._search_prefix = self._parameters['search_prefix']
        self._LOGIN_URL = 'https://' + self._host + '/api/v2/session/login'
        self._INTEL_URL = 'https://' + self._host + '/plugin/products/detect3/api/v1/intels'

        self._session = Tanium.connect_http(self._user, self._password, self._LOGIN_URL)

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        TechniqueBase.set_plugin_params(parser)

        parser.add_argument('--host', help='Tanium host', required=True)
        parser.add_argument('--user', help='Tanium API username', required=True)
        parser.add_argument('--password', help='Tanium API password', required=True)
        parser.add_argument('--search_prefix', help='Search prefix')

    def get_data_from_source(self) -> Iterable:
        """
        Gets the use-case/technique data from the source.
        :return: Iterable, yields technique, detection, applicable_to
        """
        tanium_data = self._get_all_signals()

        for signal in tanium_data:
            if 'mitreAttack' in signal.keys() and signal['mitreAttack']:
                signal_techniques = json.loads(signal['mitreAttack'])

                for t in signal_techniques['techniques']:
                    technique = t['id']
                    use_case = signal['name']
                    yield technique, use_case, None

    def _get_all_signals(self) -> dict:
        """
        Gets all signals (max 500 allowed by API) from Tanium.
        """
        headers = {'session': self._session, 'Content-Type': 'application/json'}
        params = {'limit': '500', 'name': self._search_prefix}
        r = requests.get(self._INTEL_URL, params=params, headers=headers, verify=False)
        if r.status_code == requests.codes.ok:
            return r.json()
        else:
            raise Exception(f'TechniqueTaniumSignals: get all signals failed: {r.text}')


class TechniqueElasticSecurityRules(TechniqueBase):
    """
    Class for importing Elastic Security rules with ATT&CK technique mapping.
    """
    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)

        if 'host' not in self._parameters:
            raise Exception('TechniqueElasticSecurityRules: "host" parameter is required.')
        if 'user' not in self._parameters:
            raise Exception('TechniqueElasticSecurityRules: "user" parameter is required.')
        if 'password' not in self._parameters:
            raise Exception('TechniqueElasticSecurityRules: "password" parameter is required.')

        self._host = self._parameters['host']
        self._user = self._parameters['user']
        self._password = self._parameters['password']
        self._filter = self._parameters['filter']
        self._FIND_URL = 'https://' + self._host + '/api/detection_engine/rules/_find'

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        TechniqueBase.set_plugin_params(parser)

        parser.add_argument('--host', help='Elastic Security host', required=True)
        parser.add_argument('--user', help='Elastic Security username', required=True)
        parser.add_argument('--password', help='Elastic Security password', required=True)
        parser.add_argument('--filter', help='Search filter, see Elastic documentation for more information')

    def get_data_from_source(self) -> Iterable:
        """
        Gets the use-case/technique data from the source.
        :return: Iterable, yields technique, detection, applicable_to
        """
        rule_data = self._get_all_rules()

        for rule in rule_data['data']:
            if 'threat' in rule.keys():
                for threat in rule['threat']:
                    if threat['framework'] == 'MITRE ATT&CK' and 'technique' in threat.keys():
                        for tech in threat['technique']:
                            if 'subtechnique' in tech.keys() and len(tech['subtechnique']) > 0:
                                for subtech in tech['subtechnique']:
                                    technique = subtech['id']
                                    use_case = rule['name']
                                    yield technique, use_case, None
                            else:
                                technique = tech['id']
                                use_case = rule['name']
                                yield technique, use_case, None

    def _get_all_rules(self):
        headers = {'kbn-xsrf': 'dettect', 'Content-Type': 'application/json'}
        params = {'per_page': '10000'}
        if self._filter:
            params['filter'] = self._filter
        r = requests.get(self._FIND_URL, params=params, headers=headers, auth=(self._user, self._password), verify=False)
        if r.status_code == requests.codes.ok:
            return r.json()
        else:
            raise Exception(f'TechniqueElasticSecurityRules: get all rules failed: {r.text}')


class TechniqueSuricataRules(TechniqueBase):
    """
    Import data from a Suricata rules file. It expects a metadata meta-setting containing a field with the name
    mitre_technique_id containing the ATT&CK technique ID.

    https://suricata.readthedocs.io/en/latest/rules/meta.html#metadata

    Example (taken from https://rules.emergingthreats.net/open/suricata/rules/emerging-hunting.rules):
    alert http $HOME_NET any -> $EXTERNAL_NET any (msg:"ET HUNTING Possible Phishing - Form submitted to submit-form Form Hosting";
    flow:established,to_server; http.method; content:"POST"; http.host; content:"submit-form.com"; endswith; classtype:credential-theft;
    sid:2030707; rev:2; metadata:affected_product Web_Browsers, attack_target Client_Endpoint, created_at 2020_08_20, deployment Perimeter,
    former_category HUNTING, signature_severity Critical, tag Phishing, updated_at 2020_08_20, mitre_tactic_id TA0001, mitre_tactic_name Initial_Access,
    mitre_technique_id T1566, mitre_technique_name Phishing;)
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)
        if 'file' not in self._parameters:
            raise Exception('TechniqueSuricataRules: "file" parameter is required.')

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        TechniqueBase.set_plugin_params(parser)

        parser.add_argument('--file', help='Path of the Suricate rules file to import', required=True)

    def get_data_from_source(self) -> Iterable:
        """
        Gets the use-case/technique data from the source.
        :return: Iterable, yields technique, detection, applicable_to
        """
        file = self._parameters['file']
        print(f'Reading data from "{file}"')

        from suricataparser import parse_file
        rules = parse_file(file)

        for rule in rules:
            if rule.enabled:
                for option in rule.options:
                    if option.name == 'metadata':
                        meta_data = self._convert_metadata_list_to_dict(option.value.data)
                        if 'mitre_technique_id' in meta_data.keys():
                            yield meta_data['mitre_technique_id'], rule.msg, None

    @staticmethod
    def _convert_metadata_list_to_dict(meta_data: list) -> dict:
        """
        Converts a list with "key<space>value" into a dictionary.
        """
        meta_data_dict = {}
        for item in meta_data:
            splitted = item.split(' ')
            meta_data_dict[splitted[0]] = splitted[1]
        return meta_data_dict


class TechniqueSuricataRulesSummarized(TechniqueSuricataRules):
    """
    Import data from a Suricata rules file. It expects a metadata meta-setting containing a field with the name
    mitre_technique_id containing the ATT&CK technique ID. This plugin summarizes all rules instead of naming all
    rules like in TechniqueSuricataRules plugin.

    https://suricata.readthedocs.io/en/latest/rules/meta.html#metadata

    Example (taken from https://rules.emergingthreats.net/open/suricata/rules/emerging-hunting.rules):
    alert http $HOME_NET any -> $EXTERNAL_NET any (msg:"ET HUNTING Possible Phishing - Form submitted to submit-form Form Hosting";
    flow:established,to_server; http.method; content:"POST"; http.host; content:"submit-form.com"; endswith; classtype:credential-theft;
    sid:2030707; rev:2; metadata:affected_product Web_Browsers, attack_target Client_Endpoint, created_at 2020_08_20, deployment Perimeter,
    former_category HUNTING, signature_severity Critical, tag Phishing, updated_at 2020_08_20, mitre_tactic_id TA0001, mitre_tactic_name Initial_Access,
    mitre_technique_id T1566, mitre_technique_name Phishing;)
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)

    def get_data_from_source(self) -> Iterable:
        """
        Gets the use-case/technique data from the source.
        :return: Iterable, yields technique, detection, applicable_to
        """
        file = self._parameters['file']
        print(f'Reading data from "{file}"')

        from suricataparser import parse_file
        rules = parse_file(file)

        summary = {}
        for rule in rules:
            if rule.enabled:
                for option in rule.options:
                    if option.name == 'metadata':
                        meta_data = self._convert_metadata_list_to_dict(option.value.data)
                        if 'mitre_technique_id' in meta_data.keys():
                            if meta_data['mitre_technique_id'] not in summary.keys():
                                summary[meta_data['mitre_technique_id']] = 0
                            summary[meta_data['mitre_technique_id']] += 1

        for tech_id, count in summary.items():
            yield tech_id, f'Suricata rules: #{count}', None


class TechniqueSigmaRules(TechniqueBase):
    """
    Import data from a folder with Sigma rules.
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)
        if 'folder' not in self._parameters:
            raise Exception('TechniqueSigmaRules: "folder" parameter is required.')

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        TechniqueBase.set_plugin_params(parser)

        parser.add_argument('--folder', help='Path of the folder with Sigma rules to import', required=True)

    def get_data_from_source(self) -> Iterable:
        """
        Gets the use-case/technique data from the source.
        :return: Iterable, yields technique, detection, applicable_to
        """
        folder = self._parameters['folder']

        if not os.path.isdir(folder):
            raise Exception(f'TechniqueSigmaRules: Folder does not exist: {folder}')

        from ruamel.yaml import YAML

        print(f'Reading data from "{folder}"')

        for root, _, files in os.walk(folder):
            for file in files:
                if file.endswith('.yaml') or file.endswith('.yml'):
                    filename = os.path.join(root, file)
                    yaml = YAML()
                    try:
                        with open(filename, 'r') as yaml_file:
                            yaml_content = yaml.load(yaml_file)
                    except Exception as e:
                        raise Exception(f'TechniqueSigmaRules: Failed loading YAML file "{filename}". Error: {str(e)}') from e

                    if 'tags' in yaml_content.keys():
                        for tag in yaml_content['tags']:
                            if tag.startswith('attack.t'):
                                yield tag[7:].upper(), yaml_content['title'], None


class TechniqueSplunkConfigSearches(TechniqueBase):
    """
    Import data from a Splunk config that contains saved searches (savedsearches.conf). It uses
    the action.correlationsearch.annotations attribute to get the mitre_attack techniques:

    action.correlationsearch.annotations = {"mitre_attack": ["T1560.001", "T1560"]}

    Searches that contain an action.correlationsearch.label and don't have disabled=1 are included.
    """

    def __init__(self, parameters: dict) -> None:
        super().__init__(parameters)
        if 'file' not in self._parameters:
            raise Exception('TechniqueSplunkConfigSearches: "file" parameter is required.')

    @staticmethod
    def set_plugin_params(parser: ArgumentParser) -> None:
        """
        Set command line arguments specific for the plugin
        :param parser: Argument parser
        """
        TechniqueBase.set_plugin_params(parser)

        parser.add_argument('--file', help='Path of the savedsearches config file to import', required=True)

    def get_data_from_source(self) -> Iterable:
        """
        Gets the use-case/technique data from the source.
        :return: Iterable, yields technique, detection, applicable_to
        """
        file = self._parameters['file']
        print(f'Reading data from "{file}"')

        import addonfactory_splunk_conf_parser_lib as splunk_conf_parser

        with open(file, "r") as f:
            splunk_config = splunk_conf_parser.TABConfigParser()
            splunk_config.read_file(f)

        # Read the default value for disabled attribute:
        default_disabled = '0'
        if 'default' in splunk_config.sections():
            if 'disabled' in splunk_config['default'].keys():
                default_disabled = splunk_config['default']['disabled']

        ignore_list = ['default']
        for section in splunk_config.sections():
            # Ignore some searches:
            if splunk_config[section].name in ignore_list \
               or 'action.correlationsearch.label' not in splunk_config[section].keys() \
               or 'action.correlationsearch.annotations' not in splunk_config[section].keys() \
               or ('disabled' in splunk_config[section].keys() and splunk_config[section]['disabled'] == '1') \
               or ('disabled' not in splunk_config[section].keys() and default_disabled == '1'):
                continue

            try:
                annotations = json.loads(splunk_config[section]['action.correlationsearch.annotations'])
            except Exception as e:
                print(f'Could not parse mitre_attack entry in action.correlationsearch.annotations ({str(e)}): {splunk_config[section].name}')
            else:
                if 'mitre_attack' in annotations.keys():
                    for technique in annotations['mitre_attack']:
                        yield technique, splunk_config[section].name, None
